
(** This file contains the definition of the abstract semantics of the
  example language based on separation logic. **)

Set Implicit Arguments.
Require Export FormulaExtensionsSep.
Require Export LibHeap LibList.


Global Instance : LatticeDec.t (AClosure KE).
  apply AClosure_LatDec. typeclass.
Defined.

Global Instance : PosetDec.t (FinitePowerSet.t (var * prog)).
  apply FinitePowerSet.Lat.
   typeclass.
   apply Comparable_EquivDec. typeclass.
Defined.

Definition ALocS_AValue (l : ALocS _ _) : AValueSign :=
  (⊥♯, l, ⊥♯).

Definition ALoc_AValue (l : ALoc _ _) : AValueSign :=
  ALocS_AValue (FinitePowerSet.const l).

Definition ABasic_AValue (v : ABasic) : AValueSign :=
  (v, ⊥♯, ⊥♯).

Definition AClosure_AValue x p ke : AValueSign :=
  (⊥♯, ⊥♯, FinitePowerSet.const (x, p, ke)).

Local Instance Aundef_Lat : LatticeDec.t Aundef.
  eapply LiftBot.LatDec; typeclass.
Defined.

Local Instance Aundef_Top : TopDec.t Aundef.
  apply~ LiftBot.TopDec.
Defined.

Global Instance DFormulaSign_Inhab : forall Ext (FExt : FormulaExtensionBot Ext),
    Inhab (DFormulaSign FExt).
  introv. apply~ DFormula_Inhab.
   apply~ FormulaExtension_bot_Names.
   apply~ FormulaExtension_bot_ENames.
Defined.


(** * Conditions **)

(** ** Basic Constructions **)

Section Param.

Variable Ext : Type.
(** Note that I am supposing [FormulaExtensionBot Ext] and not only
  [FormulaExtension FExt]: these additional properties greatly
  simplifies the formalisation here. **)
Variable FExt : FormulaExtensionBot Ext.

Inductive abort_ares : owhile_ares FExt → Prop :=
  | abort_ares_expr : forall Phi ar e,
    ~ e =♯ (⊥♯) →
    owhile_ares_expr (Phi, e) ⊑ ar →
    abort_ares ar
  | abort_ares_stat : forall Phi ar e,
    ~ e =♯ (⊥♯) →
    owhile_ares_stat (Phi, e) ⊑ ar →
    abort_ares ar
  | abort_ares_stat_return : forall Phi ar e,
    ~ fst (fst (EFormulaExt (proj1_sig Phi))) =♯ (⊥♯) →
    owhile_ares_stat (Phi, e) ⊑ ar →
    abort_ares ar
  .

Local Instance : Inhab (DFormulaSign (FormulaExtension_prod FormulaExtension_AValue FExt)).
  eapply DFormula_Inhab with (ext := (⊥♯, ⊥♯)).
   simpl. rewrite~ FormulaExtension_bot_Names.
   simpl. rewrite~ FormulaExtension_bot_ENames.
Defined.

End Param.

Definition abort_ast : owhile_ast → Prop.
  introv asigma. destruct asigma eqn: E;
    match goal with
    | o : owhile_ares _ |- _ =>
      exact (abort_ares _ o)
    | _ => exact False
    end.
Defined.


Definition ais_val Ext (FExt : FormulaExtension Ext) (ar : owhile_ares FExt) :=
  match ar with
  | owhile_ares_expr o =>
    let (Phi, e) := o in
    let (v, ext) := EFormulaExt (proj1_sig Phi) in
    ~ v =♯ ⊥♯
  | owhile_ares_stat o => False
  end.

Definition ais_not_basic Ext (FExt : FormulaExtension Ext) (ar : owhile_ares FExt) :=
  match ar with
  | owhile_ares_expr o =>
    let (Phi, e) := o in
    let (v, ext) := EFormulaExt (proj1_sig Phi) in
    let '(b, l, c) := v in
    ~ l =♯ (⊥♯) ∨ ~ c =♯ ⊥♯
  | owhile_ares_stat o => False
  end.

Definition ais_zero Ext (FExt : FormulaExtension Ext) (ar : owhile_ares FExt) :=
  match ar with
  | owhile_ares_expr o =>
    let (Phi, e) := o in
    let (v, ext) := EFormulaExt (proj1_sig Phi) in
    let '(b, l, c) := v in
    Sign.zero ⊑♯ b
  | owhile_ares_stat o => False
  end.

Definition ais_not_zero Ext (FExt : FormulaExtension Ext) (ar : owhile_ares FExt) :=
  match ar with
  | owhile_ares_expr o =>
    let (Phi, e) := o in
    let (v, ext) := EFormulaExt (proj1_sig Phi) in
    let '(b, l, c) := v in
    ~ b =♯ (⊥♯) ∧ ~ b =♯ Sign.zero
  | owhile_ares_stat o => False
  end.

Definition ais_loc Ext (FExt : FormulaExtension Ext) (ar : owhile_ares FExt) :=
  match ar with
  | owhile_ares_expr o =>
    let (Phi, e) := o in
    let (v, ext) := EFormulaExt (proj1_sig Phi) in
    let '(b, l, c) := v in
    ~ l =♯ ⊥♯
  | owhile_ares_stat o => False
  end.

Definition ais_not_loc Ext (FExt : FormulaExtension Ext) (ar : owhile_ares FExt) :=
  match ar with
  | owhile_ares_expr o =>
    let (Phi, e) := o in
    let (v, ext) := EFormulaExt (proj1_sig Phi) in
    let '(b, l, c) := v in
    ~ b =♯ (⊥♯) ∨ ~ c =♯ ⊥♯
  | owhile_ares_stat o => False
  end.

Definition ais_closure Ext (FExt : FormulaExtension Ext) (ar : owhile_ares FExt) :=
  match ar with
  | owhile_ares_expr o =>
    let (Phi, e) := o in
    let (v, ext) := EFormulaExt (proj1_sig Phi) in
    let '(b, l, c) := v in
    ~ c =♯ ⊥♯
  | owhile_ares_stat o => False
  end.

Definition ais_not_closure Ext (FExt : FormulaExtension Ext) (ar : owhile_ares FExt) :=
  match ar with
  | owhile_ares_expr o =>
    let (Phi, e) := o in
    let (v, ext) := EFormulaExt (proj1_sig Phi) in
    let '(b, l, c) := v in
    ~ b =♯ (⊥♯) ∨ ~ l =♯ ⊥♯
  | owhile_ares_stat o => False
  end.

Definition ais_return Ext (FExt : FormulaExtension Ext) (ar : owhile_ares FExt) :=
  match ar with
  | owhile_ares_expr o => False
  | owhile_ares_stat o =>
    let (Phi, e) := o in
    let (v, ext) := EFormulaExt (proj1_sig Phi) in
    let '(v, u) := v in
    ~ v =♯ (⊥♯)
  end.

Definition ais_normal Ext (FExt : FormulaExtension Ext) (ar : owhile_ares FExt) :=
  match ar with
  | owhile_ares_expr o => False
  | owhile_ares_stat o =>
    let (Phi, e) := o in
    let (v, ext) := EFormulaExt (proj1_sig Phi) in
    let '(v, u) := v in
    ~ u =♯ (⊥♯)
  end.


(** ** Main Construction **)

Definition owhile_acond n asigma : Prop :=
  match n, asigma with
  | name_abort_ext_expr e, asigma =>
      abort_ast asigma
  | name_abort_ext_stat s, asigma =>
      abort_ast asigma

  | name_cst c, owhile_ast_expr aSt =>
      True
  | name_random c, owhile_ast_expr aSt =>
      True
  | name_var x, owhile_ast_expr aSt =>
      True (*exists St, aState_has aSt St ∧ aindom St x*)
  | name_varUndef x, owhile_ast_expr aSt =>
      True (*exists St, aState_has aSt St ∧ anot_indom St x*)

  | name_new_obj, owhile_ast_expr Phi =>
      True
  | name_property e f, owhile_ast_expr aSt =>
      True
  | name_property_1 f, owhile_ast_property_1 ar =>
      True (*exists St l,
        aoute_is ar St (ALoc_AValue l)
        ∧ aindom_property St l f*)
  | name_property_1_no_loc f, owhile_ast_property_1 ar =>
      True (*exists St v,
        ~ v =♯ (⊥♯) ∧
        aoute_is ar St (ABasic_AValue v)*)
  | name_property_1_undef f, owhile_ast_property_1 ar =>
      True (*exists St l,
        aoute_is ar St (ALoc_AValue l)
        ∧ anot_indom_property St l f*)
  | name_op e1 op e2, owhile_ast_expr aSt =>
      True
  | name_op_1 op e2, owhile_ast_op_1 ar =>
      True (*exists St v,
        ~ v =♯ (⊥♯) ∧
        aoute_is ar St (ABasic_AValue v)*)
  | name_op_1_error op e2, owhile_ast_op_1 ar =>
      True (*exists St v,
        aval_is_loc v ∧ aoute_is ar St v*)
  | name_op_2_error op, owhile_ast_op_2 v1 ar =>
      True (*exists St v,
        aval_is_loc v ∧ aoute_is ar St v*)
  | name_add_2, owhile_ast_op_2 v1 ar =>
      True (*exists St v2,
        ~ v2 =♯ (⊥♯) ∧
        aoute_is ar St (ABasic_AValue v2)*)
  | name_sub_2, owhile_ast_op_2 v1 ar =>
      True (*exists St v2,
        ~ v2 =♯ (⊥♯) ∧
        aoute_is ar St (ABasic_AValue v2)*)
  | name_greater_2_greater, owhile_ast_op_2 v1 ar =>
      True (*exists St v2,
        aoute_is ar St (ABasic_AValue v2)
        ∧ SignMayGl v1 v2*)
  | name_greater_2_lesser_eq, owhile_ast_op_2 v1 ar =>
      True (*exists St v2,
        aoute_is ar St (ABasic_AValue v2)
        ∧ SignMayLe v1 v2*)
  | name_not e, owhile_ast_expr aSt =>
      True
  | name_not_1_true, owhile_ast_not_1 ar =>
      True (*exists St v,
        aoute_is ar St (ABasic_AValue v)
        ∧ (Sign.pos ⊑♯ v ∨ Sign.neg ⊑♯ v)*)
  | name_not_1_false, owhile_ast_not_1 ar =>
      True (*exists St v,
        aoute_is ar St (ABasic_AValue v)
        ∧ Sign.zero ⊑♯ v*)
  | name_not_1_error, owhile_ast_not_1 ar =>
      True (*exists St v,
        aval_is_loc v ∧ aoute_is ar St v*)
  | name_eq e1 e2, owhile_ast_expr aSt =>
      True
  | name_eq_1 e2, owhile_ast_eq_1 ar =>
      True (*exists St v,
        ~ v =♯ (⊥♯) ∧
        aoute_is ar St (ABasic_AValue v)*)
  | name_eq_2_basic_val_eq, owhile_ast_eq_2 ar =>
      True (*exists St v1 v2,
        aoute_param_is _ _ ar St (ABasic_AValue v1) (ABasic_AValue v2)
        ∧ ~ v1 ⊓♯ v2 =♯ (⊥♯)*)
  | name_eq_2_basic_val_neq, owhile_ast_eq_2 ar =>
      True (*exists St v1 v2,
        aoute_param_is _ _ ar St (ABasic_AValue v1) (ABasic_AValue v2)
        ∧ (~ v1 ⊑♯ Sign.zero ∨ ~ v1 ⊑♯ Sign.zero)
        (** [Sign.zero] is the only basic value with only one concretisation. **)*)
  | name_eq_2_loc_eq, owhile_ast_eq_2 ar =>
      True (*exists St l1 l2,
        aoute_param_is _ _ ar St (ALocS_AValue l1) (ALocS_AValue l2)
        ∧ ALocSMayEq St l1 l2*)
  | name_eq_2_loc_neq, owhile_ast_eq_2 ar =>
      True (*exists St l1 l2,
        aoute_param_is _ _ ar St (ALocS_AValue l1) (ALocS_AValue l2)
        ∧ ALocSMayNeq St l1 l2*)
  | name_eq_2_mistype_basic_val_loc, owhile_ast_eq_2 ar =>
      True (*exists St v1 l2,
        aoute_param_is _ _ ar St (ABasic_AValue v1) (ALocS_AValue l2)*)
  | name_eq_2_mistype_loc_basic_val, owhile_ast_eq_2 ar =>
      True (*exists St l1 v2,
        aoute_param_is _ _ ar St (ALocS_AValue l1) (ABasic_AValue v2)*)

  | name_app e1 e2, owhile_ast_expr Phi =>
      True
  | name_app_1 e2, owhile_ast_app_1 ar =>
      ais_closure ar
  | name_app_1_error e2, owhile_ast_app_1 ar =>
      ais_not_closure ar
  | name_app_2 p, owhile_ast_app_2 xp ar =>
      ais_val ar ∧ exists x,
        FinitePowerSet.const (x, p) ⊑♯ xp
  | name_app_3_return, owhile_ast_app_3 ar =>
      ais_return ar
  | name_app_3_no_return, owhile_ast_app_3 ar =>
      ais_normal ar

  | name_skip, owhile_ast_stat aSt =>
      True
  | name_seq s1 s2, owhile_ast_stat aSt =>
      True
  | name_seq_1 s2, owhile_ast_seq_1 ar =>
      True (*exists St, aouts_is ar St*)

  | name_if e s1 s2, owhile_ast_stat aSt =>
      True
  | name_if_1_true s1 s2, owhile_ast_if_1 ar =>
      True (*exists St v,
        aoute_is ar St (ABasic_AValue v)
        ∧ (Sign.pos ⊑♯ v ∨ Sign.neg ⊑♯ v)*)
  | name_if_1_false s1 s2, owhile_ast_if_1 ar =>
      True (*exists St v,
        aoute_is ar St (ABasic_AValue v)
        ∧ Sign.zero ⊑♯ v*)
  | name_if_1_error s1 s2, owhile_ast_if_1 ar =>
      True (*exists St v,
        aval_is_loc v ∧ aoute_is ar St v *)
  | name_while e s, owhile_ast_stat Phi =>
      True
  | name_while_1 e s, owhile_ast_while_1 ar =>
      ais_normal ar
  | name_while_2_true e s, owhile_ast_while_2 ar =>
      ais_not_zero ar
  | name_while_2_false e s, owhile_ast_while_2 ar =>
      ais_zero ar
  | name_while_2_error e s, owhile_ast_while_2 ar =>
      ais_not_basic ar
  | name_throw, owhile_ast_stat aSt =>
      True

  | name_property_asgn e1 f e2, owhile_ast_stat aSt =>
      True
  | name_property_asgn_1 f e2, owhile_ast_property_asgn_1 ar =>
      True (*exists St l, aoute_is ar St (ALoc_AValue l)*)
  | name_property_asgn_1_error f e2, owhile_ast_property_asgn_1 ar =>
      True (*exists St v n,
        (~ v =♯ (⊥♯) ∨ ~ n =♯ (⊥♯)) ∧ aoute_is ar St (v, ⊥♯, n)*)
  | name_property_asgn_2 f, owhile_ast_property_asgn_2 ar =>
      True (*exists St l1 v2,
        ~ l1 =♯ (⊥♯) ∧
        aoute_param_is _ _ ar St (ABasic_AValue v2) l1 *)
  | name_delete e f, owhile_ast_stat Phi =>
      True
  | name_delete_1 f, owhile_ast_delete_1 ar =>
      ais_loc ar
  | name_delete_1_error f, owhile_ast_delete_1 ar =>
      ais_not_loc ar

  | name_prog_stat s, owhile_ast_prog St =>
      True

  | _, _ =>
      False
  end.


(** ** Correctness **)

(** *** Basic Lemmae **)

Lemma abort_ares_correct : forall Ext (FExt : FormulaExtensionBot Ext) ar r ext,
  owhile_ares_gamma ar (r, ext) →
  abort_res r →
  abort_ares FExt ar.
Proof.
  introv G A. repeat inverts A as A; destruct ar; simpls; tryfalse; inverts G as G;
    try apply abort_ares_top.
   applys abort_ares_expr.
    introv A. eapply FinitePowerSet.gamma_bot with (T := error).
     apply (Gamma.monotone (t := aError_Gamma) _ _ (PosetDec.refl _ _ A) _ G).
    split; apply Poset.refl; apply Equiv.refl.
   applys abort_ares_stat.
    introv A. eapply FinitePowerSet.gamma_bot with (T := error).
     apply (Gamma.monotone (t := aError_Gamma) _ _ (PosetDec.refl _ _ A) _ G).
    split; apply Poset.refl; apply Equiv.refl.
   applys abort_ares_stat_return t1.
    introv A. lets (D&rho_out&rho_in&V_out&V_in&GT&En&(GE1&GE2)): (rm G).
     inverts GE1 as G E. eapply (@AValue_gamma_bot nat nat nat) with (ABasic_gamma := Sign.gamma).
      apply Sign.bottom_empty.
     refine (Gamma.monotone (t := AValue_Gamma _ _ _ Sign.Gamma V_in) _ _ _ _ G).
      eapply PosetDec.trans; [| apply PosetDec.refl; apply A ].
       apply PosetDec.refl. apply eq_EquivDec. apply f_equal with (f := fst) in E.
       simpl in E. rewrite~ E.
    split; apply Poset.refl; apply Equiv.refl.
Qed.

Lemma abort_ast_correct : forall asigma sigma,
  owhile_ast_gamma asigma sigma →
  abort_st sigma →
  abort_ast asigma.
Proof.
  introv G A. destruct asigma; destruct~ sigma; simpls; tryfalse;
    (applys* abort_ares_correct A || (try apply* abort_ares_param_correct)).
Qed.


(** *** Theorem **)

Lemma owhile_acond_correct : forall n asigma sigma,
  owhile_ast_gamma asigma sigma →
  owhile_cond n sigma →
  owhile_acond n asigma.
Proof.
  introv G (C1&C2). destruct n eqn: En; destruct sigma; unfolds in C1 C2; tryfalse;
    destruct asigma; unfolds in G; tryfalse; unfolds; try exact I;
    try solve [ applys* abort_ast_correct; simpl; autos~ ].
   (** cst **)
   (* Rule not yet defined *)
   (** random **)
   (* Rule not yet defined *)
   (** var **)
   (* Rule not yet defined *)
   (** var undef **)
   (* Rule not yet defined *)
   (** property **)
   (* Rule not yet defined *)
   (** property 1 **)
   (* Rule not yet defined *)
   (** property 1 no loc **)
   (* Rule not yet defined *)
   (** property 1 undef **)
   (* Rule not yet defined *)
   (** in **)
   skip. (* Rule not yet defined *)
   (** in 1 true **)
   skip. (* Rule not yet defined *)
   (** in 1 false **)
   skip. (* Rule not yet defined *)
   (** in 1 no loc **)
   skip. (* Rule not yet defined *)
   (** op **)
   (* Rule not yet defined *)
   (** op 1 **)
   (* Rule not yet defined *)
   (** op 1 error **)
   (* Rule not yet defined *)
   (** op 2 error **)
   (* Rule not yet defined *)
   (** add 2 **)
   (* Rule not yet defined *)
   (** sub 2 **)
   (* Rule not yet defined *)
   (** greater 2 greater **)
   (* Rule not yet defined *)
   (** greater 2 lesser eq **)
   (* Rule not yet defined *)
   (** not **)
   (* Rule not yet defined *)
   (** not 1 true **)
   (* Rule not yet defined *)
   (** not 1 false **)
   (* Rule not yet defined *)
   (** not 1 error **)
   (* Rule not yet defined *)
   (** eq **)
   (* Rule not yet defined *)
   (** eq 1 **)
   (* Rule not yet defined *)
   (** eq 2 basic val eq **)
   (* Rule not yet defined *)
   (** eq 2 basic val neq **)
   (* Rule not yet defined *)
   (** eq 2 loc eq **)
   (* Rule not yet defined *)
   (** eq 2 loc neq **)
   (* Rule not yet defined *)
   (** eq 2 mistype basic val loc **)
   (* Rule not yet defined *)
   (** eq 2 mistype loc basic val **)
   (* Rule not yet defined *)
   (** eq 2 closure **)
   skip. (* Rule not yet defined *)
   (** lambda **)
   skip. (* Rule not yet defined *)
   (** app 1 **)
   destruct o as [o|]; tryfalse. destruct o as [St [| |x p l]|]; tryfalse.
   destruct o0 as [[Phi er]|]; tryfalse. inverts G as G.
   lets (rho_out&rho_ini&V_out&V_in&GT&Ent&GE): (rm G). unfolds.
   destruct EFormulaExt as [[[? ?] ?] ?]. lets (GE1&GE2): (rm GE). do 2 inverts GE2 as GE2.
   lets (?&G&?): (rm GE2). introv A.
   eapply FinitePowerSet.gamma_bot with (T := (var * prog * KE)%type).
   refine (Gamma.monotone (t := FinitePowerSet.Gamma _ _) _ _ _ _ G). apply~ @PosetDec.refl.
   (** app 1 error **)
   destruct o as [o|]; tryfalse. destruct o as [St v|]; tryfalse.
   destruct o0 as [[Phi er]|]; tryfalse. inverts G as G.
   lets (rho_out&rho_ini&V_out&V_in&GT&Ent&GE): (rm G). unfolds.
   destruct EFormulaExt as [[[? ?] ?] ?]. lets (GE1&GE2): (rm GE).
   destruct v; tryfalse; do 3 inverts GE2 as GE2.
    left. introv A. false Sign.bottom_empty.
     refine (Gamma.monotone (t := Sign.Gamma) _ _ _ _ GE2). apply~ @PosetDec.refl.
    right. introv A. eapply ALocS_gamma_bot with (L := nat) (K := nat).
     refine (Gamma.monotone (t := ALocS_Gamma _ _ _) _ _ _ _ GE2).
      autos*.
      apply~ @PosetDec.refl.
   (** app 2 **)
   destruct o as [o|]; tryfalse. lets (?&IV): (rm C2). substs.
   destruct o as [St v'|]; tryfalse. lets (G1&G2): (rm G). splits.
    destruct o0 as [[Phi er]|]; tryfalse.
     unfolds. inverts G2 as G2. lets (rho_out&rho_ini&V_out&V_in&GT&Ent&GE): (rm G2).
      destruct EFormulaExt as [[[? ?] ?] ?]. lets (GE1&GE2): (rm GE).
      introv A. eapply FormulaExtension_gamma_monotone
        with (ext2 := (⊥♯ : ABasic, ⊥♯ : ALocS L K, ⊥♯ : AClosure KE, t0)) in GE2.
       inverts GE2 as GE2. refine (AValue_gamma_bot
         (L := L) (K := K) (KE := KE) (av := v') (ABasic := Sign.t) _ GE2).
       apply~ Sign.bottom_empty.
       autos~.
       apply @PosetDec.refl. split~.
    exists v. destruct t.
     apply FinitePowerSet.in_inf. introv I. simpl in I. rew_refl in I. inverts* I.
     apply TopDec.prop.
   (** app 3 return **)
   destruct o as [|[|St v'|]]; tryfalse. destruct o0 as [|[Phi er]]; tryfalse.
   unfolds. inverts G as G. lets (D&rho_out&rho_ini&V_out&V_in&GT&Ent&GE): (rm G).
   destruct EFormulaExt as [[[? ?] ?] ?]. lets (GE1&GE2): (rm GE). inverts GE1 as GE1.
   introv A. eapply (@AValue_gamma_bot nat nat nat) with (ABasic_gamma := Sign.gamma).
     apply~ Sign.bottom_empty.
   refine (Gamma.monotone (t := AValue_Gamma _ _ _ Sign.Gamma V_in) _ _ _ _ GE1).
   apply~ @PosetDec.refl.
   (** app 3 no return **)
   destruct o as [|[St| |]]; tryfalse. destruct o0 as [|[Phi er]]; tryfalse.
   unfolds. inverts G as G. lets (D&rho_out&rho_ini&V_out&V_in&GT&Ent&GE): (rm G).
   destruct EFormulaExt as [[[? ?] ?] ?]. lets (GE1&GE2): (rm GE). inverts GE1 as GE1.
   introv A. destruct a0 as [a0|]; tryfalse.
   (** skip **)
   (* Rule not yet defined *)
   (** seq **)
   (* Rule not yet defined *)
   (** seq 1 **)
   (* Rule not yet defined *)
   (** throw **)
   (* Rule not yet defined *)
   (** if **)
   (* Rule not yet defined *)
   (** if 1 true **)
   (* Rule not yet defined *)
   (** if 1 false **)
   (* Rule not yet defined *)
   (** if 1 error **)
   (* Rule not yet defined *)
   (** while 1 **)
   destruct o as [|[St| |]]; tryfalse. destruct o0 as [|[Phi er]]; tryfalse. unfolds.
   inverts G as G. lets (D&rho_out&rho_ini&V_out&V_in&GT&Ent&GE): (rm G).
   destruct EFormulaExt as [[[? ?] ?] ?]. lets (GE1&GE2): (rm GE). inverts GE1 as GE1.
   introv A. destruct a0 as [a0|]; tryfalse.
   (** while 2 true **)
   destruct o as [[St [v| |]|]|]; tryfalse. destruct o0 as [|[Phi er]]; tryfalse. unfolds.
   inverts G as G. lets (D&rho_out&rho_ini&V_out&V_in&GT&Ent&GE): (rm G).
   destruct EFormulaExt as [[[? ?] ?] ?]. lets (GE1&GE2): (rm GE). do 2 inverts GE1 as GE1.
   splits.
    introv A. inverts A. inverts GE1.
    introv A. inverts A. inverts* GE1.
   (** while 2 false **)
   destruct o as [[St [v| |]|]|]; tryfalse. destruct o0 as [|[Phi er]]; tryfalse. unfolds.
   inverts G as G. lets (D&rho_out&rho_ini&V_out&V_in&GT&Ent&GE): (rm G).
   destruct EFormulaExt as [[[? ?] ?] ?]. lets (GE1&GE2): (rm GE). do 2 inverts GE1 as GE1.
   substs. destruct a; try constructors; inverts GE1; false~.
   (** while 2 error **)
   destruct o as [o|]; tryfalse. destruct o as [v|]; tryfalse.
   destruct o0 as [|[Phi er]]; tryfalse. unfolds. inverts G as G.
   lets (D&rho_out&rho_ini&V_out&V_in&GT&Ent&GE): (rm G).
   destruct EFormulaExt as [[[? ?] ?] ?]. lets (GE1&GE2): (rm GE).
   destruct v0 as [v0|l0|c0]; tryfalse; do 2 inverts GE1 as GE1.
    left. introv A. eapply ALocS_gamma_bot with (L := nat) (K := nat).
     refine (Gamma.monotone (t := ALocS_Gamma _ _ _) _ _ _ _ GE1).
      autos*.
      apply~ @PosetDec.refl.
    right. introv A. lets (G&G'): (rm GE1).
     eapply FinitePowerSet.gamma_bot with (T := (var * prog * KE)%type).
     refine (Gamma.monotone (t := FinitePowerSet.Gamma _ _) _ _ _ _ G). apply~ @PosetDec.refl.
   (** property asgn **)
   (* Rule not yet defined *)
   (** property asgn 1 **)
   (* Rule not yet defined *)
   (** property asgn 1 error **)
   (* Rule not yet defined *)
   (** property asgn 2 **)
   (* Rule not yet defined *)
   (** delete 1 **)
   destruct o as [[St [v| |]|]|]; tryfalse. destruct o0 as [|[Phi er]]; tryfalse.
   unfolds. inverts G as G. lets (D&rho_out&rho_ini&V_out&V_in&GT&Ent&GE): (rm G).
   destruct EFormulaExt as [[[? ?] ?] ?]. lets (GE1&GE2): (rm GE). do 2 inverts GE1 as GE1.
   introv A. eapply ALocS_gamma_bot with (L := nat) (K := nat).
   refine (Gamma.monotone (t := ALocS_Gamma _ _ _) _ _ _ _ GE1).
    autos*.
    apply~ @PosetDec.refl.
   (** delete 1 error **)
   destruct o as [[St v|]|]; tryfalse. destruct o0 as [|[Phi er]]; tryfalse.
   unfolds. inverts G as G. lets (D&rho_out&rho_ini&V_out&V_in&GT&Ent&GE): (rm G).
   destruct EFormulaExt as [[[? ?] ?] ?]. lets (GE1&GE2): (rm GE).
   destruct v as [v|l|c]; tryfalse; do 2 inverts GE1 as GE1.
    left. introv A. inverts A. inverts GE1.
    right. introv A. lets (G&G'): (rm GE1).
     eapply FinitePowerSet.gamma_bot with (T := (var * prog * KE)%type).
     refine (Gamma.monotone (t := FinitePowerSet.Gamma _ _) _ _ _ _ G). apply~ @PosetDec.refl.
   (** return **)
   skip. (* Rule not yet defined *)
   (** return 1 **)
   skip. (* Rule not yet defined *)
   (** run **)
   skip. (* Rule not yet defined *)
   (** run 1 **)
   skip. (* Rule not yet defined *)
   (** stat **)
   (* Rule not yet defined *)
Qed.


(** ** Monotony **)

(** *** Small Lemmae **)

Lemma abort_ares_monotone : forall Ext (FExt : FormulaExtensionBot Ext) ar1 ar2,
  ar1 ⊑ ar2 →
  abort_ares FExt ar1 →
  abort_ares FExt ar2.
Proof.
  introv O A. inverts A as A1 A2.
   applys abort_ares_expr A1. applys~ @Poset.trans A2.
   applys abort_ares_stat A1. applys~ @Poset.trans A2.
   applys abort_ares_stat_return A1. applys~ @Poset.trans A2.
Qed.

Lemma abort_ast_monotone : forall asigma1 asigma2,
  asigma1 ⊑ asigma2 →
  abort_ast asigma1 →
  abort_ast asigma2.
Proof.
  introv O A. destruct asigma1; destruct~ asigma2; tryfalse;
    (applys abort_ares_monotone O A || applys abort_ares_param_monotone O A).
Qed.

Lemma ais_closure_monotone : forall Ext (FExt : FormulaExtension Ext)
    (o1 o2 : owhile_ares FExt),
  o1 ⊑ o2 →
  ais_closure o1 →
  ais_closure o2.
Proof.
  introv O I. destruct o1 as [[[[? [[[? ?] ?] ?]] ?] ?]|];
    destruct o2 as [[[[? [[[? ?] ?] ?]] ?] ?]|]; tryfalse.
  introv E. apply I. apply PosetDec.antisym.
   eapply PosetDec.trans; [| applys @PosetDec.refl E ].
    skip. (* Admitted *)
    apply BotDec.prop.
Qed.

Lemma ais_not_closure_monotone : forall Ext (FExt : FormulaExtension Ext)
    (o1 o2 : owhile_ares FExt),
  o1 ⊑ o2 →
  ais_not_closure o1 →
  ais_not_closure o2.
Proof.
  introv O I. destruct o1 as [[[[? [[[? ?] ?] ?]] ?] ?]|];
    destruct o2 as [[[[? [[[? ?] ?] ?]] ?] ?]|]; tryfalse.
  inverts I as I.
   left. introv E. apply I. apply Sign.order_antisym.
    eapply Sign.order_trans; [| applys Sign.order_refl E ].
     skip. (* Admitted *)
     apply Sign.bottom_is_bottom.
   right. introv E. apply I. apply PosetDec.antisym.
    eapply PosetDec.trans; [| applys @PosetDec.refl E ].
     skip. (* Admitted *)
     apply BotDec.prop.
Qed.


(** *** Monotony Theorem **)

Lemma owhile_acond_monotone : forall n (sigma sigma' : owhile_ast),
   sigma' ⊑ sigma →
   owhile_acond n sigma' →
   owhile_acond n sigma.
Proof.
  introv O C. destruct n; destruct sigma'; tryfalse; destruct sigma; tryfalse;
    unfolds; unfolds in C; trivial;
    try match goal with
    | |- abort_ast _ => try applys abort_ast_monotone O C
    | |- abort_ares _ => try applys abort_ares_monotone O C
    end.
   (** var **)
   (* To be done *)
   (** var undef **)
   (* To be done *)
   (** property 1 **)
   (* To be done *)
   (** property 1 no loc **)
   (* To be done *)
   (** property 1 undef **)
   (* To be done *)
   (** property 1 nil **)
   (* To be done *)
   (** op 1 **)
   (* To be done *)
   (** op 1 error **)
   (* To be done *)
   (** op 2 error **)
   (* To be done *)
   (** add 2 **)
   (* To be done *)
   (** sub 2 **)
   (* To be done *)
   (** greater 2 greater **)
   (* To be done *)
   (** greater 2 lesser eq **)
   (* To be done *)
   (** not 1 true **)
   (* To be done *)
   (** not 1 false **)
   (* To be done *)
   (** not 1 error **)
   (* To be done *)
   (** eq 1 **)
   (* To be done *)
   (** eq 2 basic val eq **)
   (* To be done *)
   (** eq 2 basic val neq **)
   (* To be done *)
   (** eq 2 loc eq **)
   (* To be done *)
   (** eq 2 loc neq **)
   (* To be done *)
   (** eq 2 mistype basic val loc **)
   (* To be done *)
   (** eq 2 mistype loc basic val **)
   (* To be done *)
   (** lambda **)
   (* To be done *)
   (** app **)
   (* To be done *)
   (** app 1 **)
   applys~ ais_closure_monotone C.
   (** app 1 error **)
   applys~ ais_not_closure_monotone C.
   (** app 2 **)
   skip. (* Left to be done *)
   (** app 3 return **)
   skip. (* Left to be done *)
   (** app 3 no return **)
   skip. (* Left to be done *)
   (** seq 1 **)
   (* To be done *)
   (** if 1 true **)
   (* To be done *)
   (** if 1 false **)
   (* To be done *)
   (** if 1 error **)
   (** while 1 **)
   skip. (* Left to be done *)
   (** while 2 true **)
   skip. (* Left to be done *)
   (** while 2 false **)
   skip. (* Left to be done *)
   (** while 2 error **)
   skip. (* Left to be done *)
   (** property asgn 1 **)
   (* To be done *)
   (** property asgn 1 error **)
   (* To be done *)
   (** property asgn 2 **)
   (* To be done *)
   (** delete 1 **)
   skip. (* Left to be done *)
   (** delete 1 error **)
   skip. (* Left to be done *)
   (** delete 1 nil **)
   (* To be done *)
Qed.

